/* extension.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Marco Trevisan <marco@ubuntu.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported init */

const { GObject, Shell } = imports.gi;

const Main = imports.ui.main;
const MessageTray = imports.ui.messageTray;

let _self = null;

const AppNotificationSource = GObject.registerClass(
class AppNotificationSource extends MessageTray.Source {
    _init(app) {
        super._init(app.get_name());
        this.app = app;
    }

    open() {
        this.destroy();
    }

    getIcon() {
        return this.app.icon;
    }
});

class StartupTime {
    enable() {
        this._launching = new Map();
        this._defaultLaunch = Shell.App.prototype.launch;
        this._defaultActivate = Shell.App.prototype.activate;

        Shell.App.prototype.launch = function (...args) {
            _self._monitorApp(this);
            _self._defaultLaunch.call(this, ...args);
        }

        Shell.App.prototype.activate = function (...args) {
            if (this.state === Shell.AppState.STOPPED)
                _self._monitorApp(this);
            _self._defaultActivate.call(this, ...args);
        }
    }

    _notifyUser(app, appStartupTime) {
        const msg = `${app.get_name()} was launched in ${appStartupTime}ms`;
        log(msg);

        const source = new AppNotificationSource(app);
        Main.messageTray.add(source);
        const notification = new MessageTray.Notification(source, msg);
        notification.setTransient(true);
        source.showNotification(notification);
    }

    _onAppChanged(app) {
        const now = new Date();

        if (app.state !== Shell.AppState.RUNNING || !app.get_n_windows())
            return;

        const windows = app.get_windows();
        const info = this._launching.get(app);

        info.connections?.forEach(([actor, id]) => actor.disconnect(id));
        info.connections = new Set();

        if (!windows.some(w => w.get_compositor_private()?.is_mapped())) {
            windows.forEach(w => {
                const actor = w.get_compositor_private();
                const id = actor.connect('notify::mapped', () => this._onAppChanged(app));
                info.connections.add([actor, id]);
            });

            return;
        }

        const appStartupTime = now.getTime() - info.startupTime.getTime();
        this._notifyUser(app, appStartupTime);

        this._forgetApp(app);
    }

    _forgetApp(app) {
        const info = this._launching.get(app);
        info.connections?.forEach(([actor, id]) => actor.disconnect(id));
        app.disconnect(info.stateId);
        app.disconnect(info.windowsId);
        this._launching.delete(app);
    }

    _monitorApp(app) {
        const startupTime = new Date();

        if (this._launching.has(app))
            return;

        this._launching.set(app, {
            startupTime,
            stateId: app.connect('notify::state', () => this._onAppChanged(app)),
            windowsId: app.connect('windows-changed', () => this._onAppChanged(app)),
        });

        log(`Monitoring startup time of ${app.get_name()}`);
    }

    disable() {
        this._launching.forEach(app, () => this._forgetApp(app));
        Shell.App.prototype.launch = this._defaultLaunch;
        delete this._defaultLaunch;
    }
}

function init() {
    _self = new StartupTime();
    return _self;
}
